from abc import ABC,abstractmethod
import os
from pathlib import Path
from dflow.utils import set_directory
from dflow.python import (
    PythonOPTemplate,
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    TransientError,
    FatalError,
    BigParameter,
)
from typing import (
    Any,
    Tuple,
    List,
    Set,
    Dict,
    Optional,
    Union,
)

class PrepFp(OP, ABC):
    r"""Prepares the working directories for first-principles (FP) tasks.

    A list of (same length as op_in["confs"]) working directories
    containing all files needed to start FP tasks will be
    created. The paths of the directories will be returned as
    `op["task_paths"]`. The identities of the tasks are returned as
    `op["task_names"]`.

    """

    @classmethod
    def get_input_sign(cls):
        return OPIOSign({
            "inputs" : BigParameter(object),
            "type_map": List[str],
            "confs" : Artifact(List[Path]),
            "prep_image_config" : BigParameter(dict,default={}),
            "optional_input" : BigParameter(dict,default={}),
            "optional_artifact" : Artifact(Dict[str,Path],optional=True)
        })

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            "task_names": List[str],
            "task_paths" : Artifact(List[Path]),
        })

    @abstractmethod
    def prep_task(
            self,
            conf_frame,
            inputs: Any,
            prepare_image_config: Optional[Dict] = None,
            optional_input: Optional[Dict] = None,
            optional_artifact: Optional[Dict] = None,
    ):
        r"""Define how one FP task is prepared.

        Parameters
        ----------
        conf_frame : dpdata.System
            One frame of configuration in the dpdata format.
        inputs: Any
            The class object handels all other input files of the task. 
            For example, pseudopotential file, k-point file and so on.
        prepare_image_config: Dict
            Definition of runtime parameters in the process of preparing tasks. 
        optional_input: 
            Other parameters the developers or users may need.For example:
            {
               "conf_format": "vasp/poscar"
            }
            optional_input["vasp/poscar"] is the format of the configurations that users give.
            Other keys in optional_input are defined by different developers.
        optional_artifact:
            Other files that users or developers need.The using method of this part are defined by different developers.For example, in vasp part, all the files which are given in optional_artifact will be copied to the working directory.
        """
        pass

    @OP.exec_sign_check
    def execute(
            self,
            op_in : OPIO,
    ) -> OPIO:
        r"""Execute the OP.

        Parameters
        ----------
        op_in : dict
            Input dict with components:

            - `prep_image_config` : (`dict`) It defines the parameters of the process of preparing FP tasks.
            - `inputs` : (`object`) The class object handels all other input files of the task. For example, pseudopotential file, k-point file and so on.
            - `type_map` : (`List[str]`) The list of elements.
            - `confs` : (`Artifact(List[Path])`) Configurations for the FP tasks. Stored in folders as formats which can be read by dpdata.System. 
                The format can be defined by parameter "conf_format" in "optional_input". The default format is deepmd/npy. 
            - `optional_input` : (`dict`) Other parameters the developers or users may need.For example:
                                {
                                  "conf_format": "vasp/poscar"
                                }
                                optional_input["vasp/poscar"] is the format of the configurations that users give.
                                Other keys in optional_input are defined by different developers.
            - `optional_artifact` : (` Artifact(Dict[str,Path])`) Other files that users or developers need.The using method of this part are defined by different developers.For example, in vasp part, all the files which are given in optional_artifact will be copied to the working directory.

        Returns
        -------
        op : dict 
            Output dict with components:

            - `task_names`: (`List[str]`) The name of tasks. Will be used as the identities of the tasks. The names of different tasks are different.
            - `task_paths`: (`Artifact(List[Path])`) The parepared working paths of the tasks. Contains all input files needed to start the FP. The order fo the Paths should be consistent with `op["task_names"]`
        """
        import dpdata

        inputs = op_in['inputs']
        confs = op_in['confs']
        type_map = op_in['type_map']
        prepare_image_config = op_in["prep_image_config"]
        optional_artifact = op_in["optional_artifact"]
        optional_input = op_in["optional_input"]
        try:
            conf_format = optional_input["conf_format"]
        except:
            conf_format = "deepmd/npy"

        task_names = []
        task_paths = []

        #System
        counter = 0
        # loop over list of System
        for system in confs:
            ss = dpdata.System(system, fmt=conf_format, labeled=False)
            for ff in range(ss.get_nframes()):
                nn, pp = self._exec_one_frame(counter, inputs, ss[ff], prepare_image_config, optional_input, optional_artifact)
                task_names.append(nn)
                task_paths.append(pp)
                counter += 1

        return OPIO({
            'task_names' : task_names,
            'task_paths' : task_paths,
        })


    def _exec_one_frame(
            self,
            idx,
            inputs,
            conf_frame,
            prepare_image_config = None,
            optional_input = None,
            optional_artifact = None,
    ) -> Tuple[str, Path]:
        task_name = 'task.' + '%06d' % idx
        task_path = Path(task_name)
        with set_directory(task_path,mkdir=True):
            self.prep_task(conf_frame, inputs, prepare_image_config, optional_input, optional_artifact)
        return task_name, task_path



import os
import numpy as np
import unittest

from dflow import (
    InputParameter,
    OutputParameter,
    Inputs,
    InputArtifact,
    Outputs,
    OutputArtifact,
    Workflow,
    Step,
    Steps,
    upload_artifact,
    download_artifact,
    S3Artifact,
    argo_range
)
from dflow.python import (
    PythonOPTemplate,
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    upload_packages,
)

import time, shutil, dpdata
from pathlib import Path

from context import (
        fpop,
        default_image,
        upload_python_packages,
        skip_ut_with_dflow,
        skip_ut_with_dflow_reason,
        )
from fpop.abacus import PrepAbacus,AbacusInputs
from typing import List
from constants import POSCAR_1_content,POSCAR_2_content,dump_conf_from_poscar
upload_packages.append("../fpop")
upload_packages.append("./context.py")

class TestPrepAbacus(unittest.TestCase):
    '''
    deepmd/npy format named ["data.000","data.001"].
    no optional_input or optional_artifact.
    '''
    def setUp(self):
        self.ntasks = 2
        confs = dump_conf_from_poscar("deepmd/npy",[POSCAR_1_content, POSCAR_2_content])
        self.confs = [Path(ii) for ii in confs]
        self.type_map = ['Na']

        self.source_path = Path('abacustest')
        self.source_path.mkdir(parents=True, exist_ok=True)
        (self.source_path/"INPUT").write_text('INPUT_PARAMETERS\ncalculation scf\nbasis_type lcao\n')
        (self.source_path/"KPT").write_text('here kpt')
        (self.source_path/"Na.upf").write_text('here upf')
        (self.source_path/"Na.orb").write_text('here orb')
        (self.source_path/'optional_test').write_text('here test')

        self.abacus_inputs = AbacusInputs(
            input_file=self.source_path/"INPUT",
            kpt_file=self.source_path/"KPT",
            pp_files={"Na":self.source_path/"Na.upf"},
            orb_files={"Na":self.source_path/"Na.orb"}
        )
    
    def tearDown(self):
        for ii in range(self.ntasks):
            work_path = Path("task.%06d"%ii)
            if work_path.is_dir():
                shutil.rmtree(work_path)
        if os.path.isdir(self.source_path):
            shutil.rmtree(self.source_path)
        for ii in self.confs:
            if ii.is_dir():
                shutil.rmtree(ii)

    def checkfile(self):
        tdirs = []
        for ii in range (self.ntasks):
            tdir = "task.%06d" % ii
            tdirs.append(tdir)
            print(tdir)
            self.assertTrue(Path(tdir).is_dir())
            self.assertTrue(os.path.isfile(Path(tdir)/'INPUT'))
            self.assertTrue(os.path.isfile(Path(tdir)/'KPT'))
            self.assertTrue(os.path.isfile(Path(tdir)/'STRU'))
            self.assertTrue(os.path.isfile(Path(tdir)/'Na.upf'))
            self.assertTrue(os.path.isfile(Path(tdir)/'Na.orb'))
            self.assertEqual((Path(tdir)/'INPUT').read_text().split()[0],"INPUT_PARAMETERS")      
            self.assertEqual((Path(tdir)/'KPT').read_text(),'here kpt')
            self.assertEqual((Path(tdir)/'Na.upf').read_text(),'here upf')
            self.assertEqual((Path(tdir)/'Na.orb').read_text(),'here orb')
        return tdirs

class TestPrepAbacusDpConf(TestPrepAbacus,unittest.TestCase):
    def test_prepare(self):
        op = PrepAbacus()
        out = op.execute(
            OPIO(
                {
                    "prep_image_config" : {},
                    "optional_artifact" : {"TEST": (self.source_path/'optional_test').absolute()},
                    "confs" : self.confs,
                    "inputs" : self.abacus_inputs,
                    "type_map" : self.type_map,
                }
            )
        )

        tdirs = self.checkfile()

        self.assertEqual(tdirs, out['task_names'])
        self.assertEqual(tdirs, [str(ii) for ii in out['task_paths']])

        for ii in out['task_names']:
            self.assertEqual(Path(Path(ii)/'TEST').read_text(), "here test")
     
    def testWithoutOptionalArtifact(self):
        op = PrepAbacus()
        out = op.execute(
            OPIO(
                {
                    "confs" : self.confs,
                    "inputs" : self.abacus_inputs,
                    "type_map" : self.type_map,
                }
            )
        )

        tdirs = self.checkfile()

        self.assertEqual(tdirs, out['task_names'])
        self.assertEqual(tdirs, [str(ii) for ii in out['task_paths']])

   
@unittest.skipIf(skip_ut_with_dflow, skip_ut_with_dflow_reason)
class TestPrepAbacusConf(TestPrepAbacus,unittest.TestCase):
    def test_prepare(self):
        wf = Workflow(name = "test")
        abacus = Step(
            name="PrepAbacus",
            template=PythonOPTemplate(PrepAbacus,image=default_image),
            artifacts={
                "confs":upload_artifact(self.confs),
                "optional_artifact":upload_artifact({"TEST":(self.source_path/'optional_test')}),
            },
            parameters={
                "prep_image_config" : {},
                "inputs" : self.abacus_inputs ,
                "optional_input" : {},
                "type_map" : self.type_map,
            }
        ) 
        wf.add(abacus)
        wf.submit()

        while wf.query_status() in ["Pending","Running"]:
            time.sleep(4)
        assert(wf.query_status() == 'Succeeded')
        step = wf.query_step(name="PrepAbacus")[0]
        download_artifact(step.outputs.artifacts["task_paths"])

        tdirs = self.checkfile()

        self.assertEqual(tdirs, step.outputs.parameters['task_names'].value)
        
        #check optional_artifact
        for ii in step.outputs.parameters['task_names'].value:
            self.assertEqual(Path(Path(ii)/'TEST').read_text(), "here test")

    def testWithoutOptionalParam(self):
        wf = Workflow(name = "test")
        abacus = Step(
            name="PrepAbacus",
            template=PythonOPTemplate(PrepAbacus,image=default_image),
            artifacts={
                "confs":upload_artifact(self.confs),
                "optional_artifact":upload_artifact({"TEST":(self.source_path/'optional_test')})
            },
            parameters={
                "inputs" : self.abacus_inputs ,
                "type_map" : self.type_map,
            }
        ) 
        wf.add(abacus)
        wf.submit()

        while wf.query_status() in ["Pending","Running"]:
            time.sleep(4)
        assert(wf.query_status() == 'Succeeded')
        step = wf.query_step(name="PrepAbacus")[0]
        download_artifact(step.outputs.artifacts["task_paths"])

        tdirs = self.checkfile()

        self.assertEqual(tdirs, step.outputs.parameters['task_names'].value)





import os
import numpy as np
import unittest

from dflow import (
    InputParameter,
    OutputParameter,
    Inputs,
    InputArtifact,
    Outputs,
    OutputArtifact,
    Workflow,
    Step,
    Steps,
    upload_artifact,
    download_artifact,
    S3Artifact,
    argo_range
)
from dflow.python import (
    PythonOPTemplate,
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    upload_packages,
)

import time, shutil, dpdata
from pathlib import Path

from context import (
        fpop,
        default_image,
        upload_python_packages,
        skip_ut_with_dflow,
        skip_ut_with_dflow_reason,
        )
from fpop.cp2k import PrepCp2k,Cp2kInputs
from typing import List
from constants import POSCAR_1_content,POSCAR_2_content,dump_conf_from_poscar
upload_packages.append("../fpop")
upload_packages.append("./context.py")

def check_cp2k_tasks(tcase, ntasks):
    cc = 0
    tdirs = []
    for ii in range (ntasks):
        tdir = "task.%06d" % cc
        tdirs.append(tdir)
        print(tdir)
        tcase.assertTrue(Path(tdir).is_dir())
        input = Path(tdir)/'input.inp'
        coord = Path(tdir)/'coord.xyz'
        cell = Path(tdir)/'CELL_PARAMETER'
        tcase.assertTrue(input.is_file())
        tcase.assertTrue(coord.is_file())
        tcase.assertTrue(cell.is_file())
        tcase.assertEqual(input.read_text(),'&CP2K_INPUT\n&END CP2K_INPUT')
        cc += 1
    return tdirs

class TestPrepCp2kDpConf(unittest.TestCase):
    '''
    deepmd/npy format named ["data.000","data.001"].
    no optional_input or optional_artifact.
    '''
    def setUp(self):
        self.ntasks = 2
        confs = dump_conf_from_poscar("deepmd/npy",[POSCAR_1_content, POSCAR_2_content])
        self.confs = [Path(ii) for ii in confs]
        self.inp_file = 'input.inp'
        Path(self.inp_file).write_text('&CP2K_INPUT\n&END CP2K_INPUT')
        self.type_map = ['Na']
    
    def tearDown(self):
        for ii in range(self.ntasks):
            work_path = Path("task.%06d"%ii)
            if work_path.is_dir():
                shutil.rmtree(work_path)
        if Path(self.inp_file).is_file():
            os.remove(self.inp_file)
        for ii in self.confs:
            if ii.is_dir():
                shutil.rmtree(ii)

    def test(self):
        op = PrepCp2k()
        cp2k_inputs = Cp2kInputs(self.inp_file)
        out = op.execute(
            OPIO(
                {
                    "prep_image_config" : {},
                    "optional_input" : {},
                    "confs" : self.confs,
                    "inputs" : cp2k_inputs,
                    "type_map" : self.type_map,
                }
            )
        )
        tdirs = check_cp2k_tasks(self, self.ntasks)
        self.assertEqual(tdirs, out['task_names'])
        self.assertEqual(tdirs, [str(ii) for ii in out['task_paths']])
     
    def testWithoutOptionalParameter(self):
        op = PrepCp2k()
        cp2k_inputs = Cp2kInputs(self.inp_file)
        out = op.execute(
            OPIO(
                {
                    "confs" : self.confs,
                    "inputs" : cp2k_inputs,
                    "type_map" : self.type_map,
                }
            )
        )
        tdirs = check_cp2k_tasks(self, self.ntasks)
        self.assertEqual(tdirs, out['task_names'])
        self.assertEqual(tdirs, [str(ii) for ii in out['task_paths']])
    
@unittest.skipIf(skip_ut_with_dflow, skip_ut_with_dflow_reason)
class TestPrepRunCp2kPoscarConf(unittest.TestCase):
    '''
    vasp/poscar format named ["POSCAR_1","POSCAR_2","POSCAR_3"].
    Add optional_input["conf_format"] and optional_artifact. 
    '''
    def setUp(self):
        self.ntasks = 2
        confs = dump_conf_from_poscar("deepmd/npy",[POSCAR_1_content, POSCAR_2_content])
        self.confs = [Path(ii) for ii in confs]
        self.inp_file = 'input.inp'
        Path(self.inp_file).write_text('&CP2K_INPUT\n&END CP2K_INPUT')
        self.type_map = ['Na']
        self.optional_file = Path('optional_test')
        self.optional_file.write_text('here test')

    def tearDown(self):
        for ii in range(self.ntasks):
            work_path = Path("task.%06d"%ii)
            if work_path.is_dir():
                shutil.rmtree(work_path)
        for ii in [self.inp_file,self.optional_file]:
            if Path(ii).is_file:
                os.remove(ii)
        if not self.confs:
            for ii in self.confs:
                if ii.is_file:
                    os.remove(ii)

    def test(self):
        wf = Workflow(name = "test")
        ci = Cp2kInputs(self.inp_file)
        cp2k = Step(
            name="PrepCp2k",
            template=PythonOPTemplate(PrepCp2k,image=default_image),
            artifacts={
                "confs":upload_artifact(self.confs),
                "optional_artifact":upload_artifact({"TEST":Path("optional_test")}),
            },
            parameters={
                "prep_image_config" : {},
                "inputs" : ci ,
                "optional_input" : {},
                "type_map" : self.type_map,
            }
        ) 
        wf.add(cp2k)
        wf.submit()

        while wf.query_status() in ["Pending","Running"]:
            time.sleep(4)
        assert(wf.query_status() == 'Succeeded')
        step = wf.query_step(name="PrepCp2k")[0]
        download_artifact(step.outputs.artifacts["task_paths"])

        tdirs = check_cp2k_tasks(self, self.ntasks)
        self.assertEqual(tdirs, step.outputs.parameters['task_names'].value)
        
        #check optional_artifact
        for ii in step.outputs.parameters['task_names'].value:
            self.assertEqual(Path(Path(ii)/'TEST').read_text(), "here test")

    def testWithoutOptionalParameter(self):
        wf = Workflow(name = "test")
        ci = Cp2kInputs(self.inp_file)
        cp2k = Step(
            name="PrepCp2k",
            template=PythonOPTemplate(PrepCp2k,image=default_image),
            artifacts={
                "confs":upload_artifact(self.confs),
                "optional_artifact":upload_artifact({"TEST":Path("optional_test")}),
            },
            parameters={
                "inputs" : ci ,
                "type_map" : self.type_map,
            }
        ) 
        wf.add(cp2k)
        wf.submit()

        while wf.query_status() in ["Pending","Running"]:
            time.sleep(4)
        assert(wf.query_status() == 'Succeeded')
        step = wf.query_step(name="PrepCp2k")[0]
        download_artifact(step.outputs.artifacts["task_paths"])

        tdirs = check_cp2k_tasks(self, self.ntasks)
        self.assertEqual(tdirs, step.outputs.parameters['task_names'].value)
        
        #check optional_artifact
        for ii in step.outputs.parameters['task_names'].value:
            self.assertEqual(Path(Path(ii)/'TEST').read_text(), "here test")



from mocked_ops import TestInputFiles, TestInputFiles2
from pathlib import Path
import unittest
import shutil

class TestRunInputFiles(unittest.TestCase):
    def setUp(self):
        self.task_path = Path('task')
        self.task_path.mkdir(parents=True, exist_ok=True)
        (self.task_path/'123').write_text('123')
        (self.task_path/'efg').write_text('efg')
        (self.task_path/'oo').write_text('oo')
    
    def tearDown(self):
        if Path('task').is_dir():
            shutil.rmtree('task')
    
    def test_default(self):
        fake = TestInputFiles()
        result = fake.input_files('task')
        ref = ['123', 'oo', 'efg']
        ref.sort()
        result.sort()
        self.assertEqual(result,ref)

    def test_rewrite(self):
        fake = TestInputFiles2()
        result = fake.input_files('task')
        ref = ["abc","ee"]
        ref.sort()
        result.sort()
        self.assertEqual(result,ref)



from dflow.python import (
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    upload_packages,
    FatalError,
)

upload_packages.append(__file__)

import os, json, shutil, re, pickle, glob
from pathlib import Path
from typing import Tuple, List, Optional, Dict
from context import fpop
from fpop.vasp import RunVasp
from fpop.run_fp import RunFp

class MockedRunVasp(RunVasp):
    @OP.exec_sign_check
    def execute(
            self,
            op_in : OPIO,
    ) -> OPIO:
        task_name = op_in["task_name"]
        task_path = op_in["task_path"]

        assert(op_in['task_path'].is_dir())
        assert(re.match('task.[0-9][0-9][0-9][0-9][0-9][0-9]', op_in['task_name']))
        assert(op_in['task_name'] in str(op_in['task_path']))
        for ii in ['INCAR','POSCAR','POTCAR','KPOINTS']:
            assert((op_in['task_path']/ii).is_file())

        work_dir = Path(task_name)

        cwd = os.getcwd()
        work_dir.mkdir(exist_ok=True, parents=True)
        os.chdir(work_dir)

        ifiles = glob.glob(str(task_path/'*'))
        for ii in ifiles:
            if not Path(Path(ii).name).exists():
                Path(Path(ii).name).symlink_to(ii)

        backward_dir = Path(op_in['backward_dir_name'])
        backward_dir.mkdir(exist_ok=True, parents=True)
        log = Path(Path(op_in["backward_dir_name"])/op_in["log_name"])

        fc=[]
        for ii in ['POSCAR','INCAR']:
            fc.append(Path(ii).read_text())
        log.write_text('\n'.join(fc))

        for ii in op_in["backward_list"]:
            fc = []
            fc.append(f'This is {ii} which users need in {task_name}')
            fc.append(Path('KPOINTS').read_text())
            (backward_dir / ii).write_text('\n'.join(fc))
    
        os.chdir(cwd)

        return OPIO({
            'backward_dir' : work_dir / backward_dir
        })


class TestInputFiles(RunFp):
    def run_task(
        self,
        backward_dir_name,
        log_name,
        backward_list: List[str],
        run_image_config: Optional[Dict]=None,
        optional_input: Optional[Dict]=None,
    ):
        return "mock"

class TestInputFiles2(RunFp):
    def input_files(self, task_path) -> List[str]:
        return ["abc","ee"]

    def run_task(
        self,
        backward_dir_name,
        log_name,
        backward_list: List[str],
        run_image_config: Optional[Dict]=None,
        optional_input: Optional[Dict]=None,
    ):
        return "mock"



from abc import ABC,abstractmethod
from dflow.utils import set_directory
from dflow.python import (
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    TransientError,
    FatalError,
    BigParameter,
    Parameter,
)
from dflow import (
    Workflow,
    Step,
    upload_artifact,
    download_artifact,
    InputArtifact,
    OutputArtifact,
    ShellOPTemplate
)
import os, json, shutil
from pathlib import Path
from typing import (
    Any,
    Tuple,
    List,
    Set,
    Dict,
    Optional,
    Union,
)
import numpy as np

class RunFp(OP, ABC):
    r'''Execute a first-principles (FP) task.
    A working directory named `task_name` is created. All input files
    are copied or symbol linked to directory `task_name`. The FP
    command is exectuted from directory `task_name`. 
    '''

    @classmethod
    def get_input_sign(cls):
        return OPIOSign(
            {
                "task_name": str,
                "task_path": Artifact(Path),
                "backward_list": List[str],
                "log_name": Parameter(str,default='log'),
                "backward_dir_name": Parameter(str,default='backward_dir'),
                "run_image_config": BigParameter(dict,default={}),
                "optional_artifact": Artifact(Dict[str,Path],optional=True),
                "optional_input": BigParameter(dict,default={})
            }
        )

    @classmethod
    def get_output_sign(cls):
        return OPIOSign(
            {
                "backward_dir": Artifact(Path),
            }
        )

    def input_files(self, task_path) -> List[str]:
        r'''The mandatory input files to run a FP task.
        Returns
        -------
        files: List[str]
            A list of madatory input files names.
        '''
        return os.listdir(task_path)

    @abstractmethod
    def run_task(
        self,
        backward_dir_name,
        log_name,
        backward_list: List[str],
        run_image_config: Optional[Dict]=None,
        optional_input: Optional[Dict]=None,
    ) -> str:
        r'''Defines how one FP task runs
        Parameters
        ----------
        backward_dir_name:
            The name of the directory which contains the backward files.
        log_name:
            The name of log file.
        backward_list:
            The output files the users need.
        run_image_config:
            Keyword args defined by the developer.
        optional_input:
            The parameters developers need in runtime.For example:
                                {
                                  "conf_format": "vasp/poscar"
                                }
                                optional_input["vasp/poscar"] is the format of the configurations that users give.
                                Other keys in optional_input are defined by different developers.
        
        Returns
        -------
        backward_dir_name: str
            The directory name which containers the files users need.
        '''
        pass

    @OP.exec_sign_check
    def execute(
        self,
        op_in: OPIO,
    ) -> OPIO:
        r'''Execute the OP.
        Parameters
        ----------
        op_in : dict
            Input dict with components:
            - `task_name`: (`str`) The name of task.
            - `task_path`: (`Artifact(Path)`) The path that contains all input files prepareed by `PrepFp`.
            - `backward_list`: (`List[str]`) The output files the users need.
            - `log_name`: (`str`) The name of log file.
            - `backward_dir_name`: (`str`) The name of the directory which contains the backward files.
            - `run_image_config`: (`dict`) It defines the runtime configuration of the FP task.
            - `optional_artifact` : (`Artifact(Dict[str,Path])`) Other files that users or developers need.Other files that users or developers need.The using method of this part are defined by different developers.For example, in vasp part, all the files which are given in optional_artifact will be copied to the working directory.
            - `optional_input` : (`dict`) Other parameters the developers or users may need.For example:
                                {
                                  "conf_format": "vasp/poscar"
                                }
                                optional_input["vasp/poscar"] is the format of the configurations that users give.
                                Other keys in optional_input are defined by different developers.
        Returns
        -------
            Output dict with components:
            - `backward_dir`: (`Artifact(Path)`) The directory which contains the files users need.
        Exceptions
        ----------
        TransientError
            On the failure of FP execution.
        FatalError
            When mandatory files are not found.
        '''
        run_image_config = op_in["run_image_config"]
        backward_dir_name = op_in["backward_dir_name"] 
        log_name = op_in["log_name"] 
        backward_list = op_in["backward_list"]
        optional_input = op_in["optional_input"]
        task_name = op_in["task_name"]
        task_path = op_in["task_path"]
        input_files = self.input_files(task_path)
        input_files = [(Path(task_path) / ii).resolve() for ii in input_files]
        work_dir = Path(task_name)
        opt_input_files = []
        if op_in["optional_artifact"]:
            for ss,vv in op_in["optional_artifact"].items():
                opt_input_files.append(ss)
        opt_input_files = [(Path(task_path) / ii).resolve() for ii in opt_input_files]

        with set_directory(work_dir,mkdir=True):
            # link input files
            for ii in input_files:
                if not os.path.exists(ii):
                    raise FatalError(f"cannot file file/directory {ii}")
                iname = ii.name
                Path(iname).symlink_to(ii)
            for ii in opt_input_files:
                if os.path.exists(ii):
                    iname = ii.name
                    Path(iname).symlink_to(ii)
            backward_dir_name = self.run_task(backward_dir_name,log_name,backward_list,run_image_config,optional_input)

        return OPIO(
            {
                "backward_dir": work_dir / backward_dir_name
            }
        )



from context import fpop
import os,sys,json,glob,shutil,textwrap
import dpdata
import numpy as np
import unittest
from fpop.vasp import make_kspacing_kpoints, VaspInputs
from pathlib import Path

class TestVASPInputs(unittest.TestCase):
    def setUp(self):
        Path('template.incar').write_text('foo')
        Path('POTCAR_H').write_text('bar H\n')
        Path('POTCAR_O').write_text('bar O\n')

    def tearDown(self):
        os.remove('template.incar')
        os.remove('POTCAR_H')
        os.remove('POTCAR_O')        
        if Path('POSCAR').is_file():
            os.remove('POSCAR')

    def test_make_kp (self):
        kspacing = 0.16
        gamma = False
        test_path = (Path(__file__).parent)
        all_test = glob.glob(os.path.join(test_path/'data.vasp.kp.gf', 'test.*'))
        self.assertEqual(len(all_test), 30)
        for ii in all_test :
            ss = dpdata.System(Path(ii)/'POSCAR')            
            ret=make_kspacing_kpoints(ss['cells'][0], kspacing, gamma)
            kp = [int(jj) for jj in (ret.split('\n')[3].split())]
            kp_ref = list(np.loadtxt(os.path.join(ii, 'kp.ref'), dtype = int))
            self.assertTrue(kp == kp_ref)

    def test_vasp_input_incar_potcar(self):
        iincar = 'template.incar'
        ipotcar = {'H' : 'POTCAR_H', 'O' : 'POTCAR_O'}
        vi = VaspInputs(0.16, iincar, ipotcar, True)
        self.assertEqual(vi.incar_template, 'foo')
        self.assertEqual(vi.potcars['O'], 'bar O\n')
        self.assertEqual(vi.potcars['H'], 'bar H\n')
        atom_names = ['O', 'H']
        self.assertEqual(vi.make_potcar(atom_names), 'bar O\nbar H\n')

    def test_vasp_input_kp_kgamma_true(self):
        ref = textwrap.dedent(
"""Automatic mesh
0
Gamma
10 7 7
0  0  0
"""
        )
        poscar = textwrap.dedent(
"""Foo
1
0.00 6.00 6.00
8.00 0.00 8.00
9.00 9.00 0.00
O 
1 
Selective dynamics
Cartesian
0.00 0.00 0.00 T T F
"""
        )
        Path('POSCAR').write_text(poscar)
        iincar = 'template.incar'
        ipotcar = {'H' : 'POTCAR_H', 'O' : 'POTCAR_O'}
        vi = VaspInputs(0.1, iincar, ipotcar, True)
        ss = dpdata.System('POSCAR')
        kps = vi.make_kpoints(ss['cells'][0])
        self.assertEqual(ref, kps)

        
    def test_vasp_input_kp_kgamma_false(self):
        ref = textwrap.dedent(
"""K-Points
0
Monkhorst Pack
10 7 7
0  0  0
"""
        )
        poscar = textwrap.dedent(
"""Cubic BN
1
0.00 6.00 6.00
8.00 0.00 8.00
9.00 9.00 0.00
O 
1 
Selective dynamics
Cartesian
0.00 0.00 0.00 T T F
"""
        )
        Path('POSCAR').write_text(poscar)
        iincar = 'template.incar'
        ipotcar = {'H' : 'POTCAR_H', 'O' : 'POTCAR_O'}
        vi = VaspInputs(0.1, iincar, ipotcar, False)
        ss = dpdata.System('POSCAR')
        kps = vi.make_kpoints(ss['cells'][0])
        self.assertEqual(ref, kps)



'''
If you want to do the UTs with dflow.You need export some environment variables on your machine.
export DFLOW_HOST=https://workflows.deepmodeling.com
export DFLOW_K8S_API_SERVER=https://workflows.deepmodeling.com
export DFLOW_S3_REPO_KEY=oss-bohrium
export DFLOW_S3_STORAGE_CLIENT=dflow.plugins.bohrium.TiefblueClient
export BOHRIUM_USERNAME=<bohrium-email>
export BOHRIUM_PASSWORD=<bohrium-password>
export BOHRIUM_PROJECT_ID=<bohrium-project-id>

If you don't want to do the UTs with dflow:
export SKIP_UT_WITH_DFLOW=1
'''

from dflow import config, s3_config
from dflow.plugins import bohrium
from dflow.plugins.bohrium import TiefblueClient
import os
import numpy as np
import unittest
import warnings

from dflow import (
    InputParameter,
    OutputParameter,
    Inputs,
    InputArtifact,
    Outputs,
    OutputArtifact,
    Workflow,
    Step,
    Steps,
    upload_artifact,
    download_artifact,
    S3Artifact,
    argo_range
)
from dflow.python import (
    PythonOPTemplate,
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    upload_packages,
)

import time, shutil, json, jsonpickle, dpdata
from pathlib import Path
from typing import List
    
from context import (
        fpop,
        default_image,
        upload_python_packages,
        skip_ut_with_dflow,
        skip_ut_with_dflow_reason,
        )
from fpop.vasp import PrepVasp, VaspInputs

from mocked_ops import MockedRunVasp
from fpop.preprun_fp import PrepRunFp
from constants import POSCAR_1_content,POSCAR_2_content
upload_packages.append("../fpop")
upload_packages.append("./context.py")

def dump_conf_from_poscar(
        type, 
        conf_list
        ) -> List[str] :
    for ii in range(len(conf_list)):
        Path("POSCAR_%d"%ii).write_text(conf_list[ii])
    if type == "deepmd/npy":
        confs = []
        for ii in range(len(conf_list)):
            ls = dpdata.System("POSCAR_%d"%ii, fmt="vasp/poscar")
            ls.to_deepmd_npy("data.%03d"%ii) # type: ignore
            confs.append("data.%03d")
            os.remove("POSCAR_%d"%ii)
        return confs
    elif type == "vasp/poscar":
        confs = []
        for ii in range(len(conf_list)):
            confs.append("POSCAR_%d"%ii)
        return confs
    else:
        return []

class TestMockedRunVasp(unittest.TestCase):
    def setUp(self):
        self.ntask = 6
        self.task_list = []
        for ii in range(self.ntask):
            work_path = Path("task.%06d"%ii)
            work_path.mkdir(exist_ok=True,parents=True),
            for jj in ['INCAR','POTCAR','POSCAR','KPOINTS']:
                (work_path/jj).write_text(f'{jj} in task.%06d'%ii)
            self.task_list.append(work_path)

    def check_run_vasp_output(
        self,
        task_name : str,
    ):
        cwd = os.getcwd()
        os.chdir(task_name)
        fc = []
        for ii in ['POSCAR','INCAR']:
            fc.append(Path(ii).read_text())
        self.assertEqual(fc,Path(Path('my_backward')/'my_log').read_text().strip().split('\n'))
        self.assertTrue(Path(Path('my_backward')/'POSCAR').is_file())
        self.assertTrue(Path(Path('my_backward')/'POTCAR').is_file())
        fc = []
        fc.append(f"This is POSCAR which users need in {task_name}")
        fc.append(f"KPOINTS in {task_name}")
        self.assertEqual(Path(Path('my_backward')/'POSCAR').read_text().strip().split('\n'),fc)       
        os.chdir(cwd) 

    def tearDown(self):
        for ii in range(self.ntask):
            work_path = Path("task.%06d" % ii)
            if work_path.is_dir():
                shutil.rmtree(work_path)

    def test(self):
        self.task_list_str = [str(ii) for ii in self.task_list]
        for ii in range(self.ntask):
            ip = OPIO({
                'task_name' : self.task_list_str[ii],
                'task_path' : self.task_list[ii],
                'log_name' : "my_log",
                'backward_dir_name' : "my_backward",
                "run_image_config" : {},
                "optional_artifact" : {},
                "backward_list" : ['POSCAR','POTCAR'],
                "optional_input" : {},
            })
            op = MockedRunVasp()
            out = op.execute(ip)
            self.assertEqual(out['backward_dir'] , Path(Path('task.%06d'%ii)/'my_backward'))
            self.assertTrue(Path(out['backward_dir'] / 'my_log').is_file())
            self.check_run_vasp_output(self.task_list_str[ii])

    def testWithoutOptionalParameter(self):
        self.task_list_str = [str(ii) for ii in self.task_list]
        for ii in range(self.ntask):
            ip = OPIO({
                'task_name' : self.task_list_str[ii],
                'task_path' : self.task_list[ii],
                'log_name' : "my_log",
                'backward_dir_name' : "my_backward",
                "backward_list" : ['POSCAR','POTCAR'],
            })
            op = MockedRunVasp()
            out = op.execute(ip)
            self.assertEqual(out['backward_dir'] , Path(Path('task.%06d'%ii)/'my_backward'))
            self.assertTrue(Path(out['backward_dir'] / 'my_log').is_file())
            self.check_run_vasp_output(self.task_list_str[ii])

@unittest.skipIf(skip_ut_with_dflow, skip_ut_with_dflow_reason)
class TestPrepRunVasp(unittest.TestCase):
    def setUp(self):
        self.ntasks = 2
        confs = dump_conf_from_poscar("vasp/poscar",[POSCAR_1_content, POSCAR_2_content])
        self.confs = [Path(ii) for ii in confs]
        self.incar = 'incar'
        Path(self.incar).write_text("This is INCAR")
        self.potcar = 'potcar'
        Path(self.potcar).write_text('This is POTCAR')
        self.type_map = ['Na']
        self.optional_testfile = Path('test_file')
        self.optional_testfile.write_text('This is an optional artifact')

    def check_prep_run_vasp_output(
        self,
        task_name : str,
    ):
        cwd = os.getcwd()
        os.chdir(task_name)
        self.assertTrue(Path("my_backward").is_dir())
        for ii in ['POSCAR','POTCAR','TEST','my_log']:
            self.assertTrue(Path(Path("my_backward")/ii).is_file())
        os.chdir(cwd)
    
    def tearDown(self):
        for ii in range(self.ntasks):
            work_path = Path("task.%06d" % ii)
            if work_path.is_dir():
                shutil.rmtree(work_path)
        for ii in [self.incar, self.potcar, self.optional_testfile]:
            if Path(ii).is_file():
                os.remove(ii)
        for ii in self.confs:
            if ii.is_file():
                os.remove(ii)

    def test(self):
        steps = PrepRunFp(
            "prep-run-vasp",
            PrepVasp,
            MockedRunVasp,
            prep_image = default_image,
            run_image = default_image,
        )
        vasp_inputs = VaspInputs(
            0.3,
            self.incar,
            {'Na' : self.potcar},
            True,
        )
        prep_run_step = Step(
            'prep-run-step' , 
            template = steps,
            parameters = {
                'type_map' : self.type_map,
                'inputs' : vasp_inputs,
                'optional_input' : {"conf_format" : "vasp/poscar"},
                'backward_list' : ['POSCAR','POTCAR','TEST'],
                'backward_dir_name' : 'my_backward',
                'log_name' : 'my_log',
            },
            artifacts = {
                "confs" : upload_artifact(self.confs),
                "optional_artifact" : upload_artifact({'TEST':Path("test_file")}),
            },
        )
        
        wf = Workflow(name="prerunvasp")
        wf.add(prep_run_step)
        wf.submit()

        while wf.query_status() in ["Pending", "Running"]:
            time.sleep(4)

        assert(wf.query_status() == 'Succeeded')
        step = wf.query_step(name="prep-run-step")[0]
        download_artifact(step.outputs.artifacts["backward_dirs"])
        
        task_names = ["task.%06d" % ii for ii in range(self.ntasks)]
        for ii in task_names:
            self.check_prep_run_vasp_output(ii)



import os
from dflow.config import config
from dflow.plugins.dispatcher import DispatcherExecutor

def init_executor(
    executor_dict,
):
    if executor_dict is None or config["mode"] == "debug":
        return None
    etype = executor_dict.pop("type")
    if etype == "dispatcher":
        return DispatcherExecutor(**executor_dict)
    else:
        raise RuntimeError("unknown executor type", etype)



import unittest,os
from dflow.python import OPIO,TransientError
import shutil
from pathlib import Path
from mock import mock, patch, call
from context import fpop
from fpop.cp2k import RunCp2k


class TestRunCp2k(unittest.TestCase):
    def setUp(self):
        self.cwd = os.getcwd()
        self.task_path = Path('task/path')
        self.task_path.mkdir(parents=True, exist_ok=True)
        (self.task_path / 'input.inp').write_text('here input.inp')
        (self.task_path / 'CELL_PARAMETER').write_text('here CELL_PARAMETER')
        (self.task_path / 'coord.xyz').write_text('here coord.xyz')
        (self.task_path / 'TEST1').write_text('here test1')
        (self.task_path / 'TEST2').write_text('here test2')
        self.task_name = 'task_000'
        Path(self.task_name).mkdir(parents=True, exist_ok=True)
        (Path(self.task_name)/'our_log').write_text('here log')

    def tearDown(self):
        os.chdir(self.cwd)
        if Path('task').is_dir():
            shutil.rmtree('task')
        if Path(self.task_name).is_dir():
            shutil.rmtree(self.task_name)
    
    @patch('fpop.cp2k.run_command')
    def test_success(self, mocked_run):
        mocked_run.side_effect = [ (0, 'out\n', '') ]
        op = RunCp2k()
        def new_check_run_success(obj, log_name):
            return True
        with mock.patch.object(RunCp2k, "check_run_success", new=new_check_run_success):
            out = op.execute(
                OPIO({
                    'run_image_config' :{
                        'command' : 'myCp2k',
                    },
                    'task_name' : self.task_name,
                    'task_path' : self.task_path,
                    'backward_list' : ['input.inp','TEST1'],
                    'backward_dir_name' : 'our_backward',
                    'log_name' : 'our_log',
                    'optional_input' : {},
                    'optional_artifact' : {'TEST1':Path(''),'TEST2':Path('')},
                })
            )
        work_dir = Path(self.task_name)
        # check call
        calls = [
            call(' '.join(['myCp2k', '>', 'our_log']), raise_error=False, try_bash=True, shell=True),
        ]
        mocked_run.assert_has_calls(calls)
        # check output
        self.assertEqual(out["backward_dir"], work_dir/'our_backward')
        for ii in ['input.inp', 'TEST1', 'our_log']:
            self.assertTrue(Path(Path(work_dir/'our_backward')/ ii).is_file())
        # check input files are correctly linked
        self.assertEqual((work_dir / 'input.inp').read_text(), 'here input.inp')
        self.assertEqual((work_dir / 'CELL_PARAMETER').read_text(), 'here CELL_PARAMETER')
        self.assertEqual((work_dir / 'coord.xyz').read_text(), 'here coord.xyz')
        self.assertEqual((work_dir / 'TEST1').read_text(), 'here test1')
        self.assertEqual((work_dir / 'TEST2').read_text(), 'here test2')

    @patch('fpop.cp2k.run_command')
    def test_success_without_optional_parameter(self, mocked_run):
        mocked_run.side_effect = [ (0, 'out\n', '') ]
        op = RunCp2k()
        def new_check_run_success(obj, log_name):
            return True
        with mock.patch.object(RunCp2k, "check_run_success", new=new_check_run_success):
            out = op.execute(
                OPIO({
                    'run_image_config' :{
                        'command' : 'myCp2k',
                    },
                    'task_name' : self.task_name,
                    'task_path' : self.task_path,
                    'backward_list' : ['input.inp', 'TEST1'],
                    'backward_dir_name' : 'our_backward',
                    'log_name' : 'our_log',
                    'optional_artifact' : {'TEST1':Path(''),'TEST2':Path('')},
                })
            )
        work_dir = Path(self.task_name)
        # check call
        calls = [
            call(' '.join(['myCp2k', '>', 'our_log']), raise_error=False, try_bash=True, shell=True),
        ]
        mocked_run.assert_has_calls(calls)
        # check output
        self.assertEqual(out["backward_dir"], work_dir/'our_backward')
        for ii in ['input.inp','TEST1','our_log']:
            self.assertTrue(Path(Path(work_dir/'our_backward')/ ii).is_file())
        # check input files are correctly linked
        self.assertEqual((work_dir / 'input.inp').read_text(), 'here input.inp')
        self.assertEqual((work_dir / 'CELL_PARAMETER').read_text(), 'here CELL_PARAMETER')
        self.assertEqual((work_dir / 'coord.xyz').read_text(), 'here coord.xyz')
        self.assertEqual((work_dir / 'TEST1').read_text(), 'here test1')
        self.assertEqual((work_dir / 'TEST2').read_text(), 'here test2')

    @patch('fpop.cp2k.run_command')
    def test_error(self, mocked_run):
        mocked_run.side_effect = [ (1, 'out\n', '') ]
        op = RunCp2k()
        with self.assertRaises(TransientError) as ee:
            out = op.execute(
            OPIO({
                'run_image_config' :{
                    'command' : 'myCp2k',
                },
                'task_name' : self.task_name,
                'task_path' : self.task_path,
                'backward_list' : ['input.inp', 'TEST1'],
                'optional_input' : {},
                'optional_artifact' : {'TEST1':Path(''),'TEST2':Path('')},
            })
        )
        # check call
        calls = [
            call(' '.join(['myCp2k', '>', 'log']), raise_error=False, try_bash=True, shell=True),
        ]
        mocked_run.assert_has_calls(calls)

    @patch('fpop.cp2k.run_command')
    def test_error_without_optional_parameter(self, mocked_run):
        mocked_run.side_effect = [ (1, 'out\n', '') ]
        op = RunCp2k()
        with self.assertRaises(TransientError) as ee:
            out = op.execute(
            OPIO({
                'run_image_config' :{
                    'command' : 'myCp2k',
                },
                'task_name' : self.task_name,
                'task_path' : self.task_path,
                'backward_list' : ['input.inp', 'TEST1'],
                'optional_artifact' : {'TEST1':Path(''),'TEST2':Path('')},
            })
        )
        # check call
        calls = [
            call(' '.join(['myCp2k', '>', 'log']), raise_error=False, try_bash=True, shell=True),
        ]
        mocked_run.assert_has_calls(calls)



from fpop.prep_fp import PrepFp
from fpop.run_fp import RunFp
import dpdata, sys, subprocess, os, shutil
from ase.io import read, write
from pathlib import Path
from dflow.utils import run_command
from typing import (
    Any,
    Tuple,
    List,
    Set,
    Dict,
    Optional,
    Union,
)
import numpy as np
from dargs import (
    dargs, 
    Argument, 
    Variant, 
    ArgumentEncoder,
)
from dflow.python import (
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    TransientError,
    FatalError,
    BigParameter,
)

class Cp2kInputs:
    def __init__(self, inp_file: str):
        """
        Initialize the Cp2kInputs class.

        Parameters
        ----------
        inp_file : str
            The path to the user-submitted CP2K input file.
        """
        self.inp_file_from_file(inp_file)

    @property
    def inp_template(self):
        """
        Return the template content of the input file.
        """
        return self._inp_template

    def inp_file_from_file(self, fname: str):
        """
        Read the content of the input file and store it.

        Parameters
        ----------
        fname : str
            The path to the input file.
        """
        self._inp_template = Path(fname).read_text()

    @staticmethod
    def args():
        """
        Define the arguments required by the Cp2kInputs class.
        """
        doc_inp_file = "The path to the user-submitted CP2K input file."
        return [
            Argument("inp_file", str, optional=False, doc=doc_inp_file),
        ]

class PrepCp2k(PrepFp):
    def prep_task(
            self,
            conf_frame: dpdata.System,
            inputs: Cp2kInputs,
            prepare_image_config: Optional[Dict] = None,
            optional_input: Optional[Dict] = None,
            optional_artifact: Optional[Dict] = None,
    ):
        """
        Define how one CP2K task is prepared.

        Parameters
        ----------
        conf_frame : dpdata.System
            One frame of configuration in the dpdata format.
        inputs: Cp2kInputs
            The Cp2kInputs object handles the input file of the task.
        prepare_image_config: Dict, optional
            Definition of runtime parameters in the process of preparing tasks.
        optional_input: Dict, optional
            Other parameters the developers or users may need.
        optional_artifact: Dict[str, Path], optional
            Other files that users or developers need.
        """
        # Generate POSCAR file from the configuration frame
        conf_frame.to('vasp/poscar', 'POSCAR')

        # Read the structure from the POSCAR file and write to a temporary XYZ file
        atoms = read('POSCAR')
        write('temp.xyz', atoms)

        # Read the temporary XYZ file, remove the first two lines, and write to coord.xyz
        with open('temp.xyz', 'r') as f:
            lines = f.readlines()[2:]  # Remove the first two lines
        with open('coord.xyz', 'w') as f:
            f.writelines(lines)

        # Generate CELL_PARAMETER file
        cell_params = conf_frame['cells'][0]
        with open('CELL_PARAMETER', 'w') as file:
            file.write(f"A {cell_params[0,0]:14.8f} {cell_params[0,1]:14.8f} {cell_params[0,2]:14.8f}\n")
            file.write(f"B {cell_params[1,0]:14.8f} {cell_params[1,1]:14.8f} {cell_params[1,2]:14.8f}\n")
            file.write(f"C {cell_params[2,0]:14.8f} {cell_params[2,1]:14.8f} {cell_params[2,2]:14.8f}\n")

        # Write the CP2K input file content
        Path('input.inp').write_text(inputs.inp_template)

        # Copy optional files to the working directory
        if optional_artifact:
            for file_name, file_path in optional_artifact.items():
                content = file_path.read_text()
                Path(file_name).write_text(content)


class RunCp2k(RunFp):
    def input_files(self, task_path) -> List[str]:
        """
        The mandatory input files to run a CP2K task.
        
        Returns
        -------
        files: List[str]
            A list of mandatory input file names.
        """
        return ["input.inp", "CELL_PARAMETER", "coord.xyz"]

    def run_task(
        self,
        backward_dir_name,
        log_name,
        backward_list: List[str],
        run_image_config: Optional[Dict] = None,
        optional_input: Optional[Dict] = None,
    ) -> str:
        """
        Defines how one FP task runs.

        Parameters
        ----------
        backward_dir_name : str
            The name of the directory which contains the backward files.
        log_name : str
            The name of log file.
        backward_list : List[str]
            The output files the users need. For example: ["output.log", "trajectory.xyz"]
        run_image_config : Dict, optional
            Keyword args defined by the developer. For example:
            {
              "command": "source /opt/intel/oneapi/setvars.sh && mpirun -n 64 /opt/cp2k/bin/cp2k.popt"
            }
        optional_input : Dict, optional
            The parameters developers need in runtime. For example:
            {
                "conf_format": "cp2k/input"
            }
        
        Returns
        -------
        backward_dir_name : str
            The directory name which contains the files users need.
        """
        # Get the run command
        if run_image_config:
            command = run_image_config.get("command")
            if not command:
                raise ValueError("Command not specified in run_image_config")
        else:
            raise ValueError("run_image_config is missing")
        
        # Run CP2K command and write output to log file
        command = " ".join([command, ">", log_name])
        kwargs = {"try_bash": True, "shell": True}
        if run_image_config:
            kwargs.update(run_image_config)
            kwargs.pop("command", None)
        
        # Execute command
        ret, out, err = run_command(command, raise_error=False, **kwargs)  # type: ignore
        if ret != 0:
            raise TransientError(
                "cp2k failed\n", "out msg", out, "\n", "err msg", err, "\n"
            )
        
        # Check if the task was successful
        if not self.check_run_success(log_name):
            raise TransientError(
                "cp2k failed, we could not check the exact cause. Please check the log file."
            )
        
        # Create output directory and copy log file
        os.makedirs(Path(backward_dir_name))
        shutil.copyfile(log_name, Path(backward_dir_name) / log_name)
        for ii in backward_list:
            try:
                shutil.copyfile(ii, Path(backward_dir_name) / ii)
            except:
                shutil.copytree(ii, Path(backward_dir_name) / ii)
        
        return backward_dir_name
    
    def check_run_success(self, log_name):
        """
        Check if the CP2K task ran successfully by examining the output file.

        Returns
        -------
        success : bool
            True if the task ran successfully with warnings line, False otherwise.
        """
        with open(log_name, "r") as f:
            lines = f.readlines()
        return any("The number of warnings for this run is" in line for line in lines)



from context import fpop
import os,sys,json,glob,shutil,textwrap
import dpdata
import numpy as np
import unittest
from fpop.abacus import AbacusInputs,get_pporbdpks_from_stru
from pathlib import Path
from constants import STRU1_content

class TestAbacusInputs(unittest.TestCase):
    def setUp(self):
        self.task_path = Path('abacustest')
        self.work_path = Path('abacustest/work')
        self.work_path.mkdir(parents=True, exist_ok=True)
        (self.task_path/'INPUT').write_text('INPUT_PARAMETERS\ncalculation scf\nbasis_type pw\n')
        (self.task_path/'KPT').write_text('tkpt')
        (self.task_path/'H.upf').write_text('tH.upf')
        (self.task_path/'O.upf').write_text('tO.upf')
        (self.task_path/'H.orb').write_text('tH.orb')
        (self.task_path/'O.orb').write_text('tO.orb')
        (self.task_path/'jle.orb').write_text('tjle.orb')
        (self.task_path/'model.ptg').write_bytes(bytes("tmodel.ptg",encoding="utf-8"))

        self.cwd = os.getcwd()
        os.chdir(self.work_path)

    def tearDown(self):
        os.chdir(self.cwd)
        if Path(self.task_path).is_dir():
            shutil.rmtree(self.task_path)

    def test_writefile1(self):
        abacusinput = AbacusInputs(input_file="../INPUT",
                                   pp_files= {"H": "../H.upf", "O": "../O.upf"},
                                   orb_files= {"H": "../H.orb", "O": "../O.orb"},
                                   element_mass= {"H": 1, "O": 8},
                                   kpt_file="../KPT")
        abacusinput.write_input()
        abacusinput.write_kpt()
        abacusinput.write_pporb(["H"])
        
        self.assertTrue(os.path.isfile("INPUT"))
        self.assertTrue(os.path.isfile("KPT"))
        self.assertTrue(os.path.isfile("H.upf"))
        self.assertFalse(os.path.isfile("O.upf"))
        self.assertFalse(os.path.isfile("H.orb"))
        self.assertFalse(os.path.isfile("O.orb"))

        self.assertEqual("INPUT_PARAMETERS", Path("INPUT").read_text().split("\n")[0])
        self.assertTrue("calculation scf" in Path("INPUT").read_text().split("\n"))
        self.assertTrue("basis_type pw" in Path("INPUT").read_text().split("\n"))

        self.assertEqual(Path("KPT").read_text(),"tkpt")
        self.assertEqual(Path("H.upf").read_text(),"tH.upf")
    
    def test_writefile2(self):
        abacusinput = AbacusInputs(input_file="../INPUT",
                                   pp_files= {"H": "../H.upf", "O": "../O.upf"},
                                   orb_files= {"H": "../H.orb", "O": "../O.orb"},
                                   element_mass= {"H": 1, "O": 8},
                                   kpt_file="../KPT")
        abacusinput.write_input()
        abacusinput.write_kpt()
        abacusinput.write_pporb(["H","O"])
        
        self.assertTrue(os.path.isfile("INPUT"))
        self.assertTrue(os.path.isfile("KPT"))
        self.assertTrue(os.path.isfile("H.upf"))
        self.assertTrue(os.path.isfile("O.upf"))
        self.assertFalse(os.path.isfile("H.orb"))
        self.assertFalse(os.path.isfile("O.orb"))
    
    def test_writefile3(self):
        abacusinput = AbacusInputs(input_file="../INPUT",
                                   pp_files= {"H": "../H.upf", "O": "../O.upf"},
                                   orb_files= {"H": "../H.orb", "O": "../O.orb"},
                                   element_mass= {"H": 1, "O": 8},
                                   kpt_file="../KPT")
        abacusinput.set_input("basis_type","lcao")
        self.assertTrue(abacusinput.get_input()["basis_type"],"lcao")
        abacusinput.write_input()
        abacusinput.write_kpt()
        abacusinput.write_pporb(["H","O"])
        
        self.assertTrue(os.path.isfile("INPUT"))
        self.assertTrue(os.path.isfile("KPT"))

        self.assertTrue(os.path.isfile("H.upf"))
        self.assertTrue(os.path.isfile("O.upf"))
        self.assertTrue(os.path.isfile("H.orb"))
        self.assertTrue(os.path.isfile("O.orb"))

        self.assertEqual(abacusinput.get_input()["calculation"],"scf")
        self.assertEqual(abacusinput.get_input()["basis_type"],"lcao")
        self.assertEqual(abacusinput.get_mass(["H","O"]),[1,8])
        self.assertEqual(abacusinput.get_mass(["H","O","C"]),[1,8,12.0107])
        self.assertEqual(abacusinput.get_mass(["H","O","CCC"]),[1,8,1])

        self.assertEqual("INPUT_PARAMETERS", Path("INPUT").read_text().split("\n")[0])
        self.assertTrue("calculation scf" in Path("INPUT").read_text().split("\n"))
        self.assertTrue("basis_type lcao" in Path("INPUT").read_text().split("\n"))

        self.assertEqual(Path("KPT").read_text(),"tkpt")
        self.assertEqual(Path("H.upf").read_text(),"tH.upf")
        self.assertEqual(Path("H.orb").read_text(),"tH.orb")
        self.assertEqual(Path("O.upf").read_text(),"tO.upf")
        self.assertEqual(Path("O.orb").read_text(),"tO.orb")

    def test_setpporb(self):
        abacusinput = AbacusInputs(input_file="../INPUT",
                                   pp_files= {"H": "../H.upf", "O": "../O.upf"},
                                   orb_files= {"H": "../H.orb", "O": "../O.orb"},
                                   element_mass= {"H": 1, "O": 8},
                                   kpt_file="../KPT")
        self.assertEqual(abacusinput.get_pp()["H"][1],"tH.upf")
        self.assertEqual(abacusinput.get_pp()["O"][1],"tO.upf")
        self.assertEqual(abacusinput.get_orb()["H"][1],"tH.orb")
        self.assertEqual(abacusinput.get_orb()["O"][1],"tO.orb")

        abacusinput.set_pp("H","../O.orb")
        self.assertEqual(abacusinput.get_pp()["H"][1],"tO.orb")

        abacusinput.set_orb("H","../O.upf")
        self.assertEqual(abacusinput.get_orb()["H"][1],"tO.upf")
    
    def test_deepks(self):
        abacusinput = AbacusInputs(input_file="../INPUT",
                                   pp_files= {"H": "../H.upf", "O": "../O.upf"},
                                   orb_files= {"H": "../H.orb", "O": "../O.orb"},
                                   element_mass= {"H": 1, "O": 8},
                                   deepks_descriptor="../jle.orb",
                                   deepks_model="../model.ptg",
                                   kpt_file="../KPT")
        self.assertEqual(abacusinput.get_deepks_descriptor(),("jle.orb","tjle.orb"))
        self.assertEqual(abacusinput.get_deepks_model(),("model.ptg",bytes("tmodel.ptg",encoding="utf-8")))

        abacusinput.write_deepks()
        self.assertFalse(os.path.isfile("jle.orb"))
        self.assertFalse(os.path.isfile("model.ptg"))

        abacusinput.set_input("deepks_out_labels",1)
        abacusinput.write_deepks()
        self.assertTrue(os.path.isfile("jle.orb"))
        self.assertFalse(os.path.isfile("model.ptg"))

        abacusinput.set_input("deepks_scf",1)
        abacusinput.write_deepks()
        self.assertTrue(os.path.isfile("model.ptg"))

        self.assertEqual(Path("jle.orb").read_text(),"tjle.orb")
        self.assertEqual(Path("model.ptg").read_bytes(),bytes("tmodel.ptg",encoding="utf-8"))


class TestAbacusFunctions(unittest.TestCase):
    def setUp(self):
        self.work_path = Path('abacustest')
        self.work_path.mkdir(parents=True, exist_ok=True)
        (self.work_path/'STRU').write_text(STRU1_content)

        self.cwd = os.getcwd()
        os.chdir(self.work_path)

    def tearDown(self):
        os.chdir(self.cwd)
        if Path(self.work_path).is_dir():
            shutil.rmtree(self.work_path)
    
    def test_GetPporbdpksFromStru(self):
        stru_data = get_pporbdpks_from_stru("STRU")
        self.assertTrue(stru_data["pp"],["./Ga_ONCV_PBE-1.0.upf","./As_ONCV_PBE-1.0.upf"])
        self.assertTrue(stru_data["orb"],["./Ga_gga_9au_100Ry_2s2p2d.orb","./As_gga_8au_100Ry_2s2p1d.orb"])
        self.assertTrue(stru_data["dpks"],"jle.orb")
        self.assertTrue(stru_data["labels"],["Ga","As"])
        self.assertTrue(stru_data["mass"],[69.723,74.922])










        

        


import unittest,os
from dflow.python import OPIO,TransientError
import shutil
from pathlib import Path
from mock import mock, patch, call
from context import fpop
from fpop.abacus import RunAbacus
from mocked_ops import MockedRunVasp
from constants import STRU1_content

class TestRunAbacus(unittest.TestCase):
    def setUp(self):

        self.cwd = os.getcwd()
        self.task_path = Path('abacustest')
        self.task_path.mkdir(parents=True, exist_ok=True)
        (self.task_path/"INPUT").write_text('INPUT_PARAMETERS\ncalculation scf\nbasis_type lcao\ndeepks_out_labels 1\n')
        (self.task_path/"KPT").write_text('here kpt')
        (self.task_path/"STRU").write_text(STRU1_content)
        (self.task_path/"Ga_ONCV_PBE-1.0.upf").write_text('here Ga_ONCV_PBE-1.0.upf')
        (self.task_path/"As_ONCV_PBE-1.0.upf").write_text('here As_ONCV_PBE-1.0.upf')
        (self.task_path/"Ga_gga_9au_100Ry_2s2p2d.orb").write_text('here Ga_gga_9au_100Ry_2s2p2d.orb')
        (self.task_path/"As_gga_8au_100Ry_2s2p1d.orb").write_text('here As_gga_8au_100Ry_2s2p1d.orb')
        (self.task_path/"jle.orb").write_text('here jle.orb')
        (self.task_path/'TEST1').write_text('here test1')
        (self.task_path/'TEST2').write_text('here test2')

        self.task_name = 'task_000'
        Path(self.task_name).mkdir(parents=True, exist_ok=True)
        (Path(self.task_name)/'our_log').write_text('here log')

    def tearDown(self):
        os.chdir(self.cwd)
        if os.path.isdir(self.task_path):
            shutil.rmtree(self.task_path)
        if os.path.isdir(self.task_name):
            shutil.rmtree(self.task_name)
    
    def test_inputfiles(self):
        op = RunAbacus()
        inputfiles = op.input_files(self.task_path)
        ref = ["INPUT","KPT","STRU","Ga_ONCV_PBE-1.0.upf","As_ONCV_PBE-1.0.upf","Ga_gga_9au_100Ry_2s2p2d.orb","As_gga_8au_100Ry_2s2p1d.orb","jle.orb"]
        ref.sort()
        inputfiles.sort()
        self.assertEqual(inputfiles,ref)


    @patch('fpop.abacus.run_command')
    def test_success(self, mocked_run):

        mocked_run.side_effect = [ (0, 'out\n', '') ]
        op = RunAbacus()
        def new_check_run_success(obj,logfile):
            return True
        with mock.patch.object(RunAbacus, "check_run_success", new=new_check_run_success):
            out = op.execute(
                OPIO({
                    'run_image_config' :{
                        'command' : 'myabacus',
                    },
                    'task_name' : self.task_name,
                    'task_path' : self.task_path,
                    'backward_list' : ['STRU','TEST1'],
                    'backward_dir_name' : 'our_backward',
                    'log_name' : 'our_log',
                    'optional_input' : {},
                    'optional_artifact' : {'TEST1':Path(''),'TEST2':Path('')},
                })
            )
        work_dir = Path(self.task_name)
        # check call
        calls = [
            call(' '.join(['myabacus', '>', 'our_log']), raise_error=False, try_bash=True, shell=True),
        ]
        mocked_run.assert_has_calls(calls)

        # check output
        self.assertEqual(out["backward_dir"], work_dir/'our_backward')
        for ii in ['STRU','TEST1','our_log']:
            self.assertTrue(Path(Path(work_dir/'our_backward')/ ii).is_file())
        # check input files are correctly linked
        self.assertEqual((work_dir/'STRU').read_text(), (self.task_path/"STRU").read_text())
        self.assertEqual((work_dir/'INPUT').read_text(), (self.task_path/"INPUT").read_text())
        self.assertEqual((work_dir/'KPT').read_text(), 'here kpt')
        self.assertEqual((work_dir/'Ga_ONCV_PBE-1.0.upf').read_text(), 'here Ga_ONCV_PBE-1.0.upf')
        self.assertEqual((work_dir/'As_ONCV_PBE-1.0.upf').read_text(), 'here As_ONCV_PBE-1.0.upf')
        self.assertEqual((work_dir/'Ga_gga_9au_100Ry_2s2p2d.orb').read_text(), 'here Ga_gga_9au_100Ry_2s2p2d.orb')
        self.assertEqual((work_dir/'As_gga_8au_100Ry_2s2p1d.orb').read_text(), 'here As_gga_8au_100Ry_2s2p1d.orb')
        self.assertEqual((work_dir/'jle.orb').read_text(), 'here jle.orb')
        self.assertEqual((work_dir/'TEST1').read_text(), 'here test1')
        self.assertEqual((work_dir/'TEST2').read_text(), 'here test2')

    @patch('fpop.abacus.run_command')
    def test_error(self, mocked_run):
        mocked_run.side_effect = [ (1, 'out\n', '') ]
        op = RunAbacus()
        with self.assertRaises(TransientError) as ee:
            out = op.execute(
            OPIO({
                'run_image_config' :{
                    'command' : 'myabacus',
                },
                'task_name' : self.task_name,
                'task_path' : self.task_path,
                'backward_list' : ['STRU','TEST1'],
                'optional_input' : {},
                'optional_artifact' : {'TEST1':Path(''),'TEST2':Path('')},
            })
        )
        # check call
        calls = [
            call(' '.join(['myabacus', '>', 'log']), raise_error=False, try_bash=True, shell=True),
        ]
        mocked_run.assert_has_calls(calls)

    @patch('fpop.abacus.run_command')
    def test_error_without_optional_parameter(self, mocked_run):
        mocked_run.side_effect = [ (1, 'out\n', '') ]
        op = RunAbacus()

        with self.assertRaises(TransientError) as ee:
            out = op.execute(
            OPIO({
                'run_image_config' :{
                    'command' : 'myabacus',
                },
                'task_name' : self.task_name,
                'task_path' : self.task_path,
                'backward_list' : ['STRU','TEST1'],
                'optional_artifact' : {'TEST1':Path(''),'TEST2':Path('')},
            })
        )
        # check call

        calls = [
            call(' '.join(['myabacus', '>', 'log']), raise_error=False, try_bash=True, shell=True),
        ]
        mocked_run.assert_has_calls(calls)



from fpop.prep_fp import PrepFp
from fpop.run_fp import RunFp
import sys, subprocess, os, shutil,re
from pathlib import Path
from dflow.utils import run_command
from typing import (
    Any,
    Tuple,
    List,
    Set,
    Dict,
    Optional,
    Union,
)
import numpy as np
from dflow.python import (
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    TransientError,
    FatalError,
    BigParameter,
)

MASS_DICT = {
    "H": 1.0079,
    "He": 4.0026,
    "Li": 6.941,
    "Be": 9.0122,
    "B": 10.811,
    "C": 12.0107,
    "N": 14.0067,
    "O": 15.9994,
    "F": 18.9984,
    "Ne": 20.1797,
    "Na": 22.9897,
    "Mg": 24.305,
    "Al": 26.9815,
    "Si": 28.0855,
    "P": 30.9738,
    "S": 32.065,
    "Cl": 35.453,
    "K": 39.0983,
    "Ar": 39.948,
    "Ca": 40.078,
    "Sc": 44.9559,
    "Ti": 47.867,
    "V": 50.9415,
    "Cr": 51.9961,
    "Mn": 54.938,
    "Fe": 55.845,
    "Ni": 58.6934,
    "Co": 58.9332,
    "Cu": 63.546,
    "Zn": 65.39,
    "Ga": 69.723,
    "Ge": 72.64,
    "As": 74.9216,
    "Se": 78.96,
    "Br": 79.904,
    "Kr": 83.8,
    "Rb": 85.4678,
    "Sr": 87.62,
    "Y": 88.9059,
    "Zr": 91.224,
    "Nb": 92.9064,
    "Mo": 95.94,
    "Tc": 98,
    "Ru": 101.07,
    "Rh": 102.9055,
    "Pd": 106.42,
    "Ag": 107.8682,
    "Cd": 112.411,
    "In": 114.818,
    "Sn": 118.71,
    "Sb": 121.76,
    "I": 126.9045,
    "Te": 127.6,
    "Xe": 131.293,
    "Cs": 132.9055,
    "Ba": 137.327,
    "La": 138.9055,
    "Ce": 140.116,
    "Pr": 140.9077,
    "Nd": 144.24,
    "Pm": 145,
    "Sm": 150.36,
    "Eu": 151.964,
    "Gd": 157.25,
    "Tb": 158.9253,
    "Dy": 162.5,
    "Ho": 164.9303,
    "Er": 167.259,
    "Tm": 168.9342,
    "Yb": 173.04,
    "Lu": 174.967,
    "Hf": 178.49,
    "Ta": 180.9479,
    "W": 183.84,
    "Re": 186.207,
    "Os": 190.23,
    "Ir": 192.217,
    "Pt": 195.078,
    "Au": 196.9665,
    "Hg": 200.59,
    "Tl": 204.3833,
    "Pb": 207.2,
    "Bi": 208.9804,
    "Po": 209,
    "At": 210,
    "Rn": 222,
    "Fr": 223,
    "Ra": 226,
    "Ac": 227,
    "Pa": 231.0359,
    "Th": 232.0381,
    "Np": 237,
    "U": 238.0289,
    "Am": 243,
    "Pu": 244,
    "Cm": 247,
    "Bk": 247,
    "Cf": 251,
    "Es": 252,
    "Fm": 257,
    "Md": 258,
    "No": 259,
    "Rf": 261,
    "Lr": 262,
    "Db": 262,
    "Bh": 264,
    "Sg": 266,
    "Mt": 268,
    "Rg": 272,
    "Hs": 277,
    "H": 1.0079,
    "He": 4.0026,
    "Li": 6.941,
    "Be": 9.0122,
    "B": 10.811,
    "C": 12.0107,
    "N": 14.0067,
    "O": 15.9994,
    "F": 18.9984,
    "Ne": 20.1797,
    "Na": 22.9897,
    "Mg": 24.305,
    "Al": 26.9815,
    "Si": 28.0855,
    "P": 30.9738,
    "S": 32.065,
    "Cl": 35.453,
    "K": 39.0983,
    "Ar": 39.948,
    "Ca": 40.078,
    "Sc": 44.9559,
    "Ti": 47.867,
    "V": 50.9415,
    "Cr": 51.9961,
    "Mn": 54.938,
    "Fe": 55.845,
    "Ni": 58.6934,
    "Co": 58.9332,
    "Cu": 63.546,
    "Zn": 65.39,
    "Ga": 69.723,
    "Ge": 72.64,
    "As": 74.9216,
    "Se": 78.96,
    "Br": 79.904,
    "Kr": 83.8,
    "Rb": 85.4678,
    "Sr": 87.62,
    "Y": 88.9059,
    "Zr": 91.224,
    "Nb": 92.9064,
    "Mo": 95.94,
    "Tc": 98,
    "Ru": 101.07,
    "Rh": 102.9055,
    "Pd": 106.42,
    "Ag": 107.8682,
    "Cd": 112.411,
    "In": 114.818,
    "Sn": 118.71,
    "Sb": 121.76,
    "I": 126.9045,
    "Te": 127.6,
    "Xe": 131.293,
    "Cs": 132.9055,
    "Ba": 137.327,
    "La": 138.9055,
    "Ce": 140.116,
    "Pr": 140.9077,
    "Nd": 144.24,
    "Pm": 145,
    "Sm": 150.36,
    "Eu": 151.964,
    "Gd": 157.25,
    "Tb": 158.9253,
    "Dy": 162.5,
    "Ho": 164.9303,
    "Er": 167.259,
    "Tm": 168.9342,
    "Yb": 173.04,
    "Lu": 174.967,
    "Hf": 178.49,
    "Ta": 180.9479,
    "W": 183.84,
    "Re": 186.207,
    "Os": 190.23,
    "Ir": 192.217,
    "Pt": 195.078,
    "Au": 196.9665,
    "Hg": 200.59,
    "Tl": 204.3833,
    "Pb": 207.2,
    "Bi": 208.9804,
    "Po": 209,
    "At": 210,
    "Rn": 222,
    "Fr": 223,
    "Ra": 226,
    "Ac": 227,
    "Pa": 231.0359,
    "Th": 232.0381,
    "Np": 237,
    "U": 238.0289,
    "Am": 243,
    "Pu": 244,
    "Cm": 247,
    "Bk": 247,
    "Cf": 251,
    "Es": 252,
    "Fm": 257,
    "Md": 258,
    "No": 259,
    "Rf": 261,
    "Lr": 262,
    "Db": 262,
    "Bh": 264,
    "Sg": 266,
    "Mt": 268,
    "Rg": 272,
    "Hs": 277,
}

def get_pporbdpks_from_stru(stru: str= "STRU"):
    "read the label, pp, orb, cell, coord, deepks-descriptor"
    ABACUS_STRU_KEY_WORD = [
        "ATOMIC_SPECIES",
        "NUMERICAL_ORBITAL",
        "LATTICE_CONSTANT",
        "LATTICE_VECTORS",
        "ATOMIC_POSITIONS",
        "NUMERICAL_DESCRIPTOR",
    ]
    if not os.path.isfile(stru):
        return {}
    with open(stru) as f1: lines = f1.readlines()      
    def get_block(keyname):
        block = []
        for i,line in enumerate(lines):
            if line.strip() == "": continue
            elif line.split('#')[0].strip() == keyname:
                for ij in range(i+1,len(lines)):
                    if lines[ij].strip() == "": continue
                    elif lines[ij].strip() in ABACUS_STRU_KEY_WORD:
                        return block
                    else:
                        block.append(lines[ij])
                return block
        return None
    
    atomic_species = get_block("ATOMIC_SPECIES")
    numerical_orbital = get_block("NUMERICAL_ORBITAL")
    dpks = get_block("NUMERICAL_DESCRIPTOR")
    dpks = None if dpks == None else dpks[0].strip()
    
    #read species
    pp = []
    labels = []
    mass = []
    if atomic_species:
        for line in atomic_species:
            sline = line.split()
            labels.append(sline[0])
            pp.append(sline[2])
            mass.append(float(sline[1]))
        
    #read orbital
    if numerical_orbital == None:
        orb = None
    else:
        orb = []
        for line in numerical_orbital:
            orb.append(line.split()[0])

    return {
            "labels": labels,
            "mass": mass,
            "pp":pp,
            "orb":orb,
            "dpks":dpks,}


class AbacusInputs():
    def __init__(
            self,
            input_file: Union[str,Path],
            pp_files: Dict[str, Union[str,Path]],
            element_mass: Optional[Dict[str,float]] = None,
            kpt_file: Optional[Union[str,Path]] = None,
            orb_files: Optional[Dict[str, Union[str,Path]]] = None,
            deepks_descriptor: Optional[Union[str,Path]] = None,
            deepks_model: Optional[Union[str,Path]] = None
    ):     
        """The input information of an ABACUS job except for STRU.

        Parameters
        ----------
        input_file : str
            A template INPUT file.
        element_mass : Dict[str, float]
            Specify the mass of some elements. 
            For example: {"H" : 1.0079,"O" : 15.9994}.    
        pp_files : Dict[str, str]
            The pseudopotential files for the elements. 
            For example: {"H" : "/path/to/H.upf","O" : "/path/to/O.upf"}.
        kpt_file : str, optional
            The KPT file, by default None. 
        orb_files : Dict[str, str], optional
            The numerical orbital fiels for the elements, by default None. 
            For example: {"H": "/path/to/H.orb","O": "/path/to/O.orb"}.
        deepks_descriptor : str, optional
            The deepks descriptor file, by default None.
        deepks_model : str, optional
            The deepks model file, by default None.
        """        
        self.input_file = input_file
        self._input = AbacusInputs.read_inputf(self.input_file)

        self._pp_files = self._read_dict_file(pp_files)
        self._mass = element_mass if element_mass != None else {}
        self._kpt_file = None if kpt_file == None else Path(kpt_file).read_text()
        self._orb_files = {} if orb_files == None else self._read_dict_file(orb_files)
        self._deepks_descriptor = None if deepks_descriptor == None else (os.path.split(deepks_descriptor)[1], Path(deepks_descriptor).read_text())
        self._deepks_model = None if deepks_model == None else (os.path.split(deepks_model)[1], Path(deepks_model).read_bytes())

    def _read_dict_file(self,input_dict,out_dict=None):
        # input_dict is a dict whose value is a file.
        # The filename and context will make up a tuple, which is 
        # the value of out_dict
        if not out_dict:
            out_dict = {}
        for k,v in input_dict.items():
            out_dict[k] = (os.path.split(v)[1],Path(v).read_text())
        return out_dict

    def set_input(self, key:str, value:Any):
        #if set the value to be None, can remove the key
        if value == None:
            del self._input[key.strip().lower()]
        else:
            self._input[key.strip().lower()] = value
    
    def set_mass(self, key:str, value:float):
        self._mass[key] = value
    
    def set_pp(self,key:str, value:str):
        self._read_dict_file({key:value},self._pp_files)
    
    def set_orb(self,key:str,value:str):
        self._read_dict_file({key:value},self._orb_files)
    
    def set_deepks_descriptor(self, value:str):
        self._deepks_descriptor = (os.path.split(value)[1], Path(value).read_text())
    
    def set_deepks_model(self, value:str):
        self._deepks_model = (os.path.split(value)[1], Path(value).read_bytes())

    def get_input(self):
        return self._input
    
    def get_pp(self):
        return self._pp_files
    
    def get_orb(self):
        return self._orb_files
    
    def get_deepks_descriptor(self):
        return self._deepks_descriptor
    
    def get_deepks_model(self):
        return self._deepks_model

    @staticmethod
    def read_inputf(inputf: Union[str,Path]) -> dict:
        """Read INPUT and transfer to a dict.

        Parameters
        ----------
        inputf : str
            INPUT file name

        Returns
        -------
        dict[str,str]
            all input parameters
        """  
        input_context = {}
        with open(inputf) as f1: input_lines = f1.readlines()
        readinput = False
        for i,iline in enumerate(input_lines):
            if iline.strip() == 'INPUT_PARAMETERS':
                readinput = True
            elif iline.strip() == '' or iline.strip()[0] in ['#']:
                continue
            elif readinput:
                sline =re.split('[ \t]',iline.split("#")[0].strip(),maxsplit=1)
                if len(sline) == 2:
                    input_context[sline[0].lower().strip()] = sline[1].strip() 
        return input_context    

    def write_input(self,inputf :str = "INPUT"):
        with open(inputf,'w') as f1:
            f1.write("INPUT_PARAMETERS\n")
            for k,v in self._input.items():
                f1.write("%s %s\n" % (str(k),str(v)))
    
    def write_kpt(self,kptf = "KPT"):
        if self._kpt_file:
            Path(kptf).write_text(self._kpt_file)

    def write_pporb(self,element_list : List[str]):
        """Based on element list, write the pp/orb files, and return a list of the filename. 

        Parameters
        ----------
        element_list : List[str]
            a list of element name

        Returns
        -------
        List[List]
            a list of the list of pp files, and orbital files 
        """  
        need_orb = False
        if self._input.get("basis_type","pw").lower() in ["lcao","lcao_in_pw"]:
            need_orb = True
        pp,orb = [],[]
        for ielement in element_list:
            if ielement in self._pp_files:
                Path(self._pp_files[ielement][0]).write_text(self._pp_files[ielement][1])
                pp.append(self._pp_files[ielement][0])
            if need_orb and ielement in self._orb_files:
                Path(self._orb_files[ielement][0]).write_text(self._orb_files[ielement][1]) 
                orb.append(self._orb_files[ielement][0])

        if not orb: 
            orb = None

        return [pp,orb]     

    def write_deepks(self):
        """Check if INPUT is a deepks job, if yes, will return the deepks descriptor file name, 
        else will return None.

        Returns
        -------
        str
            deepks descriptor file name or None.
        """
        need_descriptor =  need_model = False  
        if self._input.get("deepks_out_labels",False):
            need_descriptor = True
        if self._input.get("deepks_scf",False):
            need_descriptor = True
            need_model = True 

        if need_descriptor:
            assert(self._deepks_descriptor != None)
            descriptor_file = self._deepks_descriptor[0]
            Path(descriptor_file).write_text(self._deepks_descriptor[1])
        else:
            descriptor_file = None

        if need_model:
            assert(self._deepks_model != None)
            Path(self._deepks_model[0]).write_bytes(self._deepks_model[1])

        return descriptor_file
        
    def get_mass(self,element_list: List[str]) -> List[float]:
        """Get the mass of elements.
        If the element is not specified in self._mass, this funciton will firstly search it 
        in a standard element-mass dictionary. And if the element is also not found in the
        dictionary, the mass will be set to 1.0. 

        Parameters
        ----------
        element_list : List[str]
            element name

        Returns
        -------
        List[float]
            the mass of each element
        """ 
        mass = []
        for i in element_list:
            mass.append(self._mass.get(i,MASS_DICT.get(i,1.0)))
        return mass

class PrepAbacus(PrepFp):
    def prep_task(
            self,
            conf_frame,
            inputs: AbacusInputs,
            prepare_image_config: Optional[Dict] = None,
            optional_input: Optional[Dict] = None,
            optional_artifact: Optional[Dict] = None,
    ):
        r"""Define how one Abacus task is prepared.

        Parameters
        ----------
        conf_frame : dpdata.System
            One frame of configuration in the dpdata format.
        inputs: AbacusInputs
            The AbacusInputs object handels all other input files of the task.
        prepare_image_config: Dict
            Definition of runtime parameters in the process of preparing tasks. 
        optional_input: 
            Other parameters the developers or users may need.
        optional_artifact
            Other files that users or developers need. 
        """

        element_list = conf_frame['atom_names']
        pp, orb = inputs.write_pporb(element_list)
        dpks = inputs.write_deepks()
        mass = inputs.get_mass(element_list)
        conf_frame.to('abacus/stru', 'STRU', pp_file=pp,numerical_orbital=orb,numerical_descriptor=dpks,mass=mass)
        
        inputs.write_input("INPUT")
        inputs.write_kpt("KPT")

        if optional_artifact:
            for file_name, file_path in optional_artifact.items():
                content = file_path.read_text()
                Path(file_name).write_text(content)
       

class RunAbacus(RunFp):
    def input_files(self,task_path) -> List[str]:
        r'''The mandatory input files to run an abacus task.
        Returns
        -------
        files: List[str]
            A list of madatory input files names.
        '''
        
        cwd = os.getcwd()
        os.chdir(task_path)

        files = ["INPUT","STRU"]
        if os.path.isfile("KPT"):
            files.append("KPT")
            
        files_tmp = []
        #read STRU
        stru_data = get_pporbdpks_from_stru("STRU")
        if stru_data != None:
            orb_files = stru_data["orb"]
            pp_files = stru_data["pp"]
            dpks_descriptor = stru_data["dpks"]

            files_tmp += pp_files
            if orb_files: files_tmp += orb_files
            if dpks_descriptor: files_tmp += [dpks_descriptor]

        #read INPUT
        input = AbacusInputs.read_inputf("INPUT")
        if "deepks_model" in input: files_tmp += [input["deepks_model"]]

        for ii in files_tmp:
            if os.path.isfile(ii):
                files.append(ii)
            else:
                print("ERROR: file %s is not found" % ii)

        os.chdir(cwd)

        return files

    def run_task(
        self,
        backward_dir_name,
        log_name,
        backward_list: List[str],
        run_image_config: Optional[Dict]=None,
        optional_input: Optional[Dict]=None,
    ) -> str:
        r'''Defines how one FP task runs
        Parameters
        ----------
        backward_dir_name:
            The name of the directory which contains the backward files.
        log_name:
            The name of log file.
        backward_list:
            The output files the users need.
        run_image_config:
            Keyword args defined by the developer.For example:
            {
              "command": "source /opt/intel/oneapi/setvars.sh && mpirun -n 16 abacus"
            }
        optional_input:
            The parameters developers need in runtime.
        
        Returns
        -------
        backward_dir_name: str
            The directory name which containers the files users need.
        '''
        
        if run_image_config:
            command = run_image_config["command"]
        else:
            command = "abacus"
        # run abacus
        command = " ".join([command, ">", log_name])
        kwargs = {"try_bash": True, "shell": True}
        if run_image_config:
            kwargs.update(run_image_config)
            kwargs.pop("command", None)
        ret, out, err = run_command(command, raise_error=False, **kwargs) # type: ignore
        if ret != 0:
            raise TransientError(
                "abacus failed\n", "out msg", out, "\n", "err msg", err, "\n"
            )
        if not self.check_run_success(log_name):
            raise TransientError(
                "abacus failed , we could not check the exact cause . Please check log file ."
            )
        os.makedirs(Path(backward_dir_name))
        shutil.copyfile(log_name,Path(backward_dir_name)/log_name)
        for ii in backward_list:
            try:
                shutil.copyfile(ii,Path(backward_dir_name)/ii)
            except:
                shutil.copytree(ii,Path(backward_dir_name)/ii)
        return backward_dir_name

    def check_run_success(self,log_name):
        with open(log_name,"r") as f:
            lines = f.readlines()
        if "SEE INFORMATION IN" in lines[-1]:
            return True
        else:
            return False



from context import fpop
import os,sys,json,glob,shutil,textwrap
import dpdata
import numpy as np
import unittest
from fpop.cp2k import Cp2kInputs
from pathlib import Path

class TestCP2KInputs(unittest.TestCase):
    def setUp(self):
        Path('template.inp').write_text('&GLOBAL\n  PROJECT foo\n&END GLOBAL\n')

    def tearDown(self):
        os.remove('template.inp')

    def test_cp2k_input(self):
        iinp_file = 'template.inp'
        ci = Cp2kInputs(iinp_file)
        self.assertEqual(ci.inp_template, '&GLOBAL\n  PROJECT foo\n&END GLOBAL\n')
